########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.service` --- Utilities for implementing :class:`~.Client` services
#################################################################################

Protocol extensions or in general support for parts of the XMPP protocol are
implemented using :class:`Service` classes, or rather, classes which use the
:class:`Meta` metaclass.

Both of these are provided in this module. To reduce the boilerplate required
to develop services, :ref:`decorators <api-aioxmpp.service-decorators>` are
provided which can be used to easily register coroutines and functions as
stanza handlers, filters and others.

.. autoclass:: Service

.. _api-aioxmpp.service-decorators:

Decorators and Descriptors
==========================

These decorators provide special functionality when used on methods of
:class:`Service` subclasses.

.. note::

   These decorators work only on methods declared on :class:`Service`
   subclasses, as their functionality are implemented in cooperation with the
   :class:`Meta` metaclass and :class:`Service` itself.

.. note::

    These decorators and the descriptors (see below) are initialised in the
    order in which they are declared at the class. In many cases, this does
    not matter, but there are some corner cases.

    For example: Suppose you have a class like this:

    .. code-block:: python

        class FooService(aioxmpp.service.Service):
            feature = aioxmpp.disco.register_feature(
                "some:namespace"
            )

            @aioxmpp.servie.depsignal(aioxmpp.DiscoServer, "on_info_changed")
            def handle_on_info_changed(self):
                pass

    In this case, the ``handle_on_info_changed`` method is not invoked during
    startup of the ``FooService``. In this case however:

    .. code-block:: python

        class FooService(aioxmpp.service.Service):
            @aioxmpp.servie.depsignal(aioxmpp.DiscoServer, "on_info_changed")
            def handle_on_info_changed(self):
                pass

            feature = aioxmpp.disco.register_feature(
                "some:namespace"
            )

    The ``handle_on_info_changed`` *is* invoked during startup of the
    ``FooService`` because the ``some:namespace`` feature is registered
    *after* the signal is connected.

    .. versionchanged:: 0.9

        This behaviour was introduced in version 0.9.

   When using a  descriptor and a :func:`depsignal`
   connected to :meth:`.DiscoServer.on_info_changed`: if the
   :class:`.disco.register_feature` is declared *before* the :func:`depsignal`,
   the signal handler will not be invoked for that specific feature because
   it is registered before the signal handler is connected).

.. autodecorator:: iq_handler

.. autodecorator:: message_handler

.. autodecorator:: presence_handler

.. autodecorator:: inbound_message_filter()

.. autodecorator:: inbound_presence_filter()

.. autodecorator:: outbound_message_filter()

.. autodecorator:: outbound_presence_filter()

.. autodecorator:: depsignal

.. autodecorator:: depfilter

.. autodecorator:: attrsignal

.. seealso::

   :class:`~.disco.register_feature`
      For a descriptor (see below) which allows to register a Service Discovery
      feature when the service is instantiated.

   :class:`~.disco.mount_as_node`
      For a descriptor (see below) which allows to register a Service Discovery
      node when the service is instantiated.

   :class:`~.pep.register_pep_node`
      For a descriptor (see below) which allows to register a PEP node
      including notification features.

Test functions
--------------

.. autofunction:: is_iq_handler

.. autofunction:: is_message_handler

.. autofunction:: is_presence_handler

.. autofunction:: is_inbound_message_filter

.. autofunction:: is_inbound_presence_filter

.. autofunction:: is_outbound_message_filter

.. autofunction:: is_outbound_presence_filter

.. autofunction:: is_depsignal_handler

.. autofunction:: is_depfilter_handler

.. autofunction:: is_attrsignal_handler

Creating your own decorators
----------------------------

Sometimes, when you create your own service, it makes sense to create own
decorators which depending services can use to make easy use of some features
of your service.

.. note::

   Remember that it isn’t necessary to create custom decorators to simply
   connect a method to a signal exposed by another service. Users of that
   service should be using :func:`depsignal` instead.

The key part is the :class:`HandlerSpec` object. It specifies the effect the
decorator has on initialisation and shutdown of the service. To add a
:class:`HandlerSpec` to a decorated method, use :func:`add_handler_spec` in the
implementation of your decorator.

.. autoclass:: HandlerSpec(key, is_unique=True, require_deps=[])

.. autofunction:: add_handler_spec

Creating your own descriptors
-----------------------------

Sometimes a decorator is not the right tool for the job, because with what you
attempt to achieve, there’s simply no relationship to a method.

In this case, subclassing :class:`Descriptor` is the way to go. It provides an
abstract base class implementing a :term:`descriptor`. Using a
:class:`Descriptor` subclass, you can create objects for each individual
service instance using the descriptor, including cleanup.

.. autoclass:: Descriptor

Metaclass
=========

.. autoclass:: Meta()


"""

import abc
import asyncio
import collections
import contextlib
import copy
import logging
import warnings
import weakref

import aioxmpp.callbacks
import aioxmpp.stream


def automake_magic_attr(obj):
    obj._aioxmpp_service_handlers = getattr(
        obj, "_aioxmpp_service_handlers", set()
    )
    return obj._aioxmpp_service_handlers


def get_magic_attr(obj):
    return obj._aioxmpp_service_handlers


def has_magic_attr(obj):
    return hasattr(
        obj, "_aioxmpp_service_handlers"
    )


class Descriptor(metaclass=abc.ABCMeta):
    """
    Abstract base class for resource managing descriptors on :class:`Service`
    classes.

    While resources such as callback slots can easily be managed with
    decorators (see above), because they are inherently related to the method
    they use, others cannot. A :class:`Descriptor` provides a method to
    initialise a context manager. The context manager is entered when the
    service is initialised and left when the service is shut down, thus
    providing a way for the :class:`Descriptor` to manage the resource
    associated with it.

    The result from entering the context manager is accessible by reading the
    attribute the descriptor is bound to.

    Subclasses must implement the following:

    .. automethod:: init_cm

    .. autoattribute:: value_type

    Subclasses may override the following to modify the default behaviour:

    .. autoattribute:: required_dependencies

    .. automethod:: add_to_stack

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._data = weakref.WeakKeyDictionary()

    @property
    def required_dependencies(self):
        """
        Iterable of services which must be declared as dependencies on a class
        using this descriptor.

        The default implementation returns an empty list.
        """
        return []

    @abc.abstractmethod
    def init_cm(self, instance):
        """
        Create and return a :term:`context manager`.

        :param instance: The service instance for which the CM is used.
        :return: A context manager managing the resource.

        The context manager is responsible for acquiring, initialising,
        destructing and releasing the resource managed by this descriptor.

        The returned context manager is not stored anywhere in the descriptor,
        it is the responsibility of the caller to register it appropriately.
        """

    def add_to_stack(self, instance, stack):
        """
        Get the context manager for the service `instance` and push it to the
        context manager `stack`.

        :param instance: The service to get the context manager for.
        :type instance: :class:`Service`
        :param stack: The context manager stack to push the CM onto.
        :type stack: :class:`contextlib.ExitStack`
        :return: The object returned by the context manager on enter.

        If a context manager has already been created for `instance`, it is
        re-used.

        On subsequent calls to :meth:`__get__` for the given `instance`, the
        return value of this method will be returned, that is, the value
        obtained from entering the context.
        """

        cm = self.init_cm(instance)
        obj = stack.enter_context(cm)
        self._data[instance] = cm, obj
        return obj

    def __get__(self, instance, owner):
        if instance is None:
            return self
        try:
            cm, obj = self._data[instance]
        except KeyError:
            raise AttributeError(
                "resource manager descriptor has not been initialised"
            )
        return obj

    @abc.abstractproperty
    def value_type(self):
        """
        The type of the value of the descriptor, once it is being accessed
        as an object attribute.

        .. versionadded:: 0.9
        """


class DependencyGraphNode:
    def __init__(self):
        self.class_ = None

    def __repr__(self):
        return "DependencyGraphNode({!r})".format(self.class_)


class DependencyGraph:

    def __init__(self, edges=None, nodes=None):
        if edges is None:
            edges = []
        if nodes is None:
            nodes = []
        self._edges = set(edges)
        self._nodes = set(nodes)

    def __deepcopy__(self, memo):
        return DependencyGraph(self._edges, self._nodes)

    def add_node(self, node):
        self._nodes.add(node)

    def add_edge(self, from_, to):
        self._edges.add((from_, to))

    def toposort(self):
        edges = collections.defaultdict(lambda: set())
        for from_, to in self._edges:
            edges[from_].add(to)

        sorted_ = []
        marked = set()
        done = set()

        def visit(node):
            if node in marked:
                raise ValueError("dependency loop in service definitions")
            if node in done:
                return
            done.add(node)
            marked.add(node)
            for dep in edges[node]:
                visit(dep)
            marked.remove(node)
            sorted_.append(node)

        for node in self._nodes:
            visit(node)

        return sorted_


class Meta(abc.ABCMeta):
    """
    The metaclass for services. The :class:`Service` class uses it and in
    general you should just inherit from :class:`Service` and define the
    dependency attributes as needed.

    Only use :class:`Meta` explicitely if you know what you are doing,
    and you most likely do not. :class:`Meta` is internal API and may
    change at any point.

    Services have dependencies. A :class:`Meta` instance (i.e. a service class)
    can declare dependencies using the following attributes.

    .. attribute:: ORDER_BEFORE

       An iterable of :class:`Service` classes before which the class which is
       currently being declared needs to be instanciated.

       Thus, any service which occurs in :attr:`ORDER_BEFORE` will be
       instanciated *after* this class (if at all). Think of it as "*this*
       class is ordered *before* the classes in this attribute".

       .. versionadded:: 0.3

    .. attribute:: SERVICE_BEFORE

       Before 0.3, this was the name of the :attr:`ORDER_BEFORE` attribute. It
       is still supported, but use emits a :data:`DeprecationWarning`. It must
       not be mixed with :attr:`ORDER_BEFORE` or :attr:`ORDER_AFTER` on a class
       declaration, or the declaration will raise :class:`ValueError`.

       .. deprecated:: 0.3

          Support for this attribute will be removed in 1.0; starting with 1.0,
          using this attribute will raise a :class:`TypeError` on class
          declaration and a :class:`AttributeError` when accessing it on a
          class or instance.

    .. attribute:: ORDER_AFTER

       An iterable of :class:`Service` classes which will be instanciated
       *before* the class which is being declraed.

       Classes which are declared in this attribute are always instanciated
       before this class is instantiated. Think of it as "*this* class is
       ordered *after* the classes in this attribute".

       .. versionadded:: 0.3

    .. attribute:: SERVICE_AFTER

       Before 0.3, this was the name of the :attr:`ORDER_AFTER` attribute. It
       is still supported, but use emits a :data:`DeprecationWarning`. It must
       not be mixed with :attr:`ORDER_BEFORE` or :attr:`ORDER_AFTER` on a class
       declaration, or the declaration will raise :class:`ValueError`.

       .. deprecated:: 0.3

          See :attr:`SERVICE_BEFORE` for details on the deprecation cycle.

    Further, the following attributes are generated:

    .. attribute:: PATCHED_ORDER_AFTER

       An iterable of :class:`Service` classes. This includes all
       classes in :attr:`ORDER_AFTER` and all classes which specify the class
       in :attr:`ORDER_BEFORE`.

       This is primarily used internally to handle :attr:`ORDER_BEFORE` when
       summoning services.

       It is an error to manually define :attr:`PATCHED_ORDER_AFTER` in a class
       definition, doing so will raise a :class:`TypeError`.

       .. versionadded:: 0.9

    .. attribute:: _DEPGRAPH_NODE

       An internal token used for topological ordering. Consider this
       name reserved by the metaclass.

       It is an error to manually define :attr:`_DEPGRAPH_NODE` in a class
       definition, doing so will raise a :class:`TypeError`.

       .. versionadded:: 0.9

    .. versionchanged:: 0.9

       The :attr:`ORDER_AFTER` and :attr:`ORDER_BEFORE` attribtes do not
       change after class creation. In earlier versions they contained
       the transitive completion of the dependency relation.

    Dependency relationships must not have cycles; a cycle results in a
    :class:`ValueError` when the class causing the cycle is declared.

    .. note::

      Subclassing instances of :class:`Meta` is forbidden. Trying to do so
      will raise a :class:`TypeError`

      .. versionchanged:: 0.9

    Example::

        class Foo(metaclass=service.Meta):
            pass

        class Bar(metaclass=service.Meta):
            ORDER_BEFORE = [Foo]

        class Baz(metaclass=service.Meta):
            ORDER_BEFORE = [Bar]

        class Fourth(metaclass=service.Meta):
            ORDER_BEFORE = [Bar]

    ``Baz`` and ``Fourth`` will be instanciated before ``Bar`` and ``Bar`` will
    be instanciated before ``Foo``. There is no dependency relationship between
    ``Baz`` and ``Fourth``.

    """

    __dependency_graph = DependencyGraph()
    __service_order = {}

    def __new__(mcls, name, bases, namespace, inherit_dependencies=True):
        if "SERVICE_BEFORE" in namespace or "SERVICE_AFTER" in namespace:
            if "ORDER_BEFORE" in namespace or "ORDER_AFTER" in namespace:
                raise ValueError("declaration mixes old and new ordering "
                                 "attribute names (SERVICE_* vs. ORDER_*)")
            warnings.warn(
                "SERVICE_BEFORE/AFTER used on class; use ORDER_BEFORE/AFTER",
                DeprecationWarning)
            try:
                namespace["ORDER_BEFORE"] = namespace.pop("SERVICE_BEFORE")
            except KeyError:
                pass
            try:
                namespace["ORDER_AFTER"] = namespace.pop("SERVICE_AFTER")
            except KeyError:
                pass

        if "PATCHED_ORDER_AFTER" in namespace:
            raise TypeError(
                "PATCHED_ORDER_AFTER must not be defined manually. "
                "it is supplied automatically by the metaclass."
            )

        if "_DEPGRAPH_NODE" in namespace:
            raise TypeError(
                "_DEPGRAPH_NODE must not be defined manually. "
                "it is supplied automatically by the metaclass."
            )

        for base in bases:
            if isinstance(base, Meta) and base is not Service:
                raise TypeError(
                    "subclassing services is prohibited."
                )

        for base in bases:
            if hasattr(base, "SERVICE_HANDLERS") and base.SERVICE_HANDLERS:
                raise TypeError(
                    "inheritance from service class with handlers is forbidden"
                )

        namespace["ORDER_BEFORE"] = frozenset(
            namespace.get("ORDER_BEFORE", ()))
        namespace["ORDER_AFTER"] = frozenset(
            namespace.get("ORDER_AFTER", ()))
        namespace["PATCHED_ORDER_AFTER"] = namespace["ORDER_AFTER"]

        new_deps = copy.deepcopy(mcls.__dependency_graph)

        depgraph_node = namespace["_DEPGRAPH_NODE"] = DependencyGraphNode()
        new_deps.add_node(depgraph_node)
        for cls in namespace["ORDER_AFTER"]:
            new_deps.add_edge(depgraph_node, cls._DEPGRAPH_NODE)
        for cls in namespace["ORDER_BEFORE"]:
            new_deps.add_edge(cls._DEPGRAPH_NODE, depgraph_node)
        sorted_ = new_deps.toposort()
        mcls.__dependency_graph = new_deps
        mcls.__service_order = dict(
            (node, i) for i, node in enumerate(sorted_)
        )

        SERVICE_HANDLERS = []
        existing_handlers = set()

        for attr_name, attr_value in namespace.items():
            if has_magic_attr(attr_value):
                new_handlers = get_magic_attr(attr_value)

                unique_handlers = {
                    spec.key
                    for spec in new_handlers
                    if spec.is_unique
                }

                conflicting = unique_handlers & existing_handlers
                if conflicting:
                    key = next(iter(conflicting))
                    obj = next(iter(
                        obj
                        for obj_key, obj in SERVICE_HANDLERS
                        if obj_key == key
                    ))

                    raise TypeError(
                        "handler conflict between {!r} and {!r}: "
                        "both want to use {!r}".format(
                            obj,
                            attr_value,
                            key,
                        )
                    )

                existing_handlers |= unique_handlers

                for spec in new_handlers:
                    missing = spec.require_deps - namespace["ORDER_AFTER"]
                    if missing:
                        raise TypeError(
                            "decorator requires dependency {!r} "
                            "but it is not declared".format(
                                next(iter(missing))
                            )
                        )

                    SERVICE_HANDLERS.append(
                        (spec.key, attr_value)
                    )

            elif isinstance(attr_value, Descriptor):
                missing = set(attr_value.required_dependencies) - \
                    namespace["ORDER_AFTER"]
                if missing:
                    raise TypeError(
                        "descriptor requires dependency {!r} "
                        "but it is not declared".format(
                            next(iter(missing)),
                        )
                    )

                SERVICE_HANDLERS.append(attr_value)

        namespace["SERVICE_HANDLERS"] = tuple(SERVICE_HANDLERS)

        return super().__new__(mcls, name, bases, namespace)

    def __init__(self, name, bases, namespace, inherit_dependencies=True):
        super().__init__(name, bases, namespace)
        for cls in self.ORDER_BEFORE:
            cls.PATCHED_ORDER_AFTER |= frozenset([self])
        self._DEPGRAPH_NODE.class_ = self

    def __prepare__(*args, **kwargs):
        return collections.OrderedDict()

    @property
    def SERVICE_BEFORE(self):
        return self.ORDER_BEFORE

    @property
    def SERVICE_AFTER(self):
        return self.ORDER_AFTER

    def __lt__(self, other):
        return (self.__service_order[self._DEPGRAPH_NODE] <
                self.__service_order[other._DEPGRAPH_NODE])

    def __le__(self, other):
        return (self.__service_order[self._DEPGRAPH_NODE] <=
                self.__service_order[other._DEPGRAPH_NODE])


class Service(metaclass=Meta):
    """
    A :class:`Service` is used to implement XMPP or XEP protocol parts, on top
    of the more or less fixed stanza handling implemented in
    :mod:`aioxmpp.node` and :mod:`aioxmpp.stream`.

    :class:`Service` is a base class which can be used by extension developers
    to implement support for custom or standardized protocol extensions. Some
    of the features for which :mod:`aioxmpp` has support are also implemented
    using :class:`Service` subclasses.

    `client` must be a :class:`~.Client` to which the service will be attached.
    The `client` cannot be changed later, for the sake of simplicity.

    `logger_base` may be a :class:`logging.Logger` instance or :data:`None`. If
    it is :data:`None`, a logger is automatically created, by taking the fully
    qualified name of the :class:`Service` subclass which is being
    instanciated. Otherwise, the logger is passed to :meth:`derive_logger` and
    the result is used as value for the :attr:`logger` attribute.

    To implement your own service, derive from :class:`Service`. If your
    service depends on other services (such as :mod:`aioxmpp.pubsub` or
    :mod:`aioxmpp.disco`), these dependencies *must* be declared as documented
    in the service meta class :class:`Meta`.

    To stay forward compatible, accept arbitrary keyword arguments and pass
    them down to :class:`Service`. As it is not possible to directly pass
    arguments to :class:`Service`\ s on construction (due to the way
    :meth:`aioxmpp.Client.summon` works), there is no need for you
    to introduce custom arguments, and thus there should be no conflicts.

    .. note::

       Inheritance from classes which subclass :class:`Service` is forbidden.

       .. versionchanged:: 0.9

    .. autoattribute:: client

    .. autoattribute:: dependencies

    .. automethod:: derive_logger

    .. automethod:: shutdown
    """

    def __init__(self, client, *, logger_base=None, dependencies={}):
        if logger_base is None:
            self.logger = logging.getLogger(".".join([
                type(self).__module__, type(self).__qualname__
            ]))
        else:
            self.logger = self.derive_logger(logger_base)

        super().__init__()
        self.__context = contextlib.ExitStack()
        self.__client = client
        self.__dependencies = dependencies

        for item in self.SERVICE_HANDLERS:
            if isinstance(item, Descriptor):
                item.add_to_stack(self, self.__context)
            else:
                (handler_cm, additional_args), obj = item
                self.__context.enter_context(
                    handler_cm(
                        self,
                        self.__client.stream,
                        obj.__get__(self, type(self)),
                        *additional_args
                    )
                )

    def derive_logger(self, logger):
        """
        Return a child of `logger` specific for this instance. This is called
        after :attr:`client` has been set, from the constructor.

        The child name is calculated by the default implementation in a way
        specific for aioxmpp services; it is not meant to be used by
        non-:mod:`aioxmpp` classes; do not rely on the way how the child name
        is calculated.
        """
        parts = type(self).__module__.split(".")[1:]
        if parts[-1] == "service" and len(parts) > 1:
            del parts[-1]

        return logger.getChild(".".join(
            parts+[type(self).__qualname__]
        ))

    @property
    def client(self):
        """
        The client to which the :class:`Service` is bound. This attribute is
        read-only.

        If the service has been shut down using :meth:`shutdown`, this reads as
        :data:`None`.
        """
        return self.__client

    @property
    def dependencies(self):
        """
        When the service is instantiated through
        :meth:`~.Client.summon`, this attribute holds a mapping which maps the
        service classes contained in the :attr:`~.Meta.ORDER_AFTER` attribute
        to the respective instances related to the :attr:`client`.

        This is the preferred way to obtain dependencies specified via
        :attr:`~.Meta.ORDER_AFTER`.
        """
        return self.__dependencies

    @asyncio.coroutine
    def _shutdown(self):
        """
        Actual implementation of the shut down process.

        This *must* be called using super from inheriting classes after their
        own shutdown procedure. Inheriting classes *must* override this method
        instead of :meth:`shutdown`.
        """

    @asyncio.coroutine
    def shutdown(self):
        """
        Close the service and wait for it to completely shut down.

        Some services which are still running may depend on this service. In
        that case, the service may refuse to shut down instead of shutting
        down, by raising a :class:`RuntimeError` exception.

        .. note::

           Developers creating subclasses of :class:`Service` to implement
           services should not override this method. Instead, they should
           override the :meth:`_shutdown` method.

        """
        yield from self._shutdown()
        self.__context.close()
        self.__client = None


class HandlerSpec(collections.namedtuple(
        "HandlerSpec",
        [
            "is_unique",
            "key",
            "require_deps",
        ])):
    """
    Specification of the effects of the decorator at initalisation and shutdown
    time.

    :param key: Context manager and arguments pair.
    :type key: pair
    :param is_unique: Whether multiple identical `key` values are allowed on a
                      single class.
    :type is_unique: :class:`bool`
    :param require_deps: Dependent services which are required for the
                         decorator to work.
    :type require_deps: iterable of :class:`Service` classes

    During initialisation of the :class:`Service` which has a method using a
    given handler spec, the first part of the `key` pair is called with the
    service instance as first, the client :class:`StanzaStream` as second and
    the bound method as third argument. The second part of the `key` is
    unpacked as additional positional arguments.

    The result of the call must be a context manager, which is immediately
    entered. On shutdown, the context manager is exited.

    An example use would be the following handler spec::

      HandlerSpec(
          (func, (IQType.GET, some_payload_class)),
          is_unique=True,
      )

    where ``func`` is a context manager which takes a service instance, a
    stanza stream, a bound method as well as an IQ type and a payload class. On
    enter, the context manager would register the method it received as third
    argument on the stanza stream (second argument) as handler for the given IQ
    type and payload class (fourth and fifth arguments).

    If `is_unique` is true and several methods have :class:`HandlerSpec`
    objects with the same `key`, :class:`TypeError` is raised at class
    definition time.

    If at class definition time any of the dependent classes in `require_deps`
    are not declared using the order attributes (see :class:`Meta`), a
    :class:`TypeError` is raised.
    """

    def __new__(cls, key, is_unique=True, require_deps=()):
        return super().__new__(cls, is_unique, key, frozenset(require_deps))


def add_handler_spec(f, handler_spec):
    """
    Attach a handler specification (see :class:`HandlerSpec`) to a function.

    :param f: Function to attach the handler specification to.
    :param handler_spec: Handler specification to attach to the function.
    :type handler_spec: :class:`HandlerSpec`

    This uses a private attribute, whose exact name is an implementation
    detail. The `handler_spec` is stored in a :class:`set` bound to the
    attribute.
    """
    automake_magic_attr(f).add(handler_spec)


def _apply_iq_handler(instance, stream, func, type_, payload_cls):
    return aioxmpp.stream.iq_handler(stream, type_, payload_cls, func)


def _apply_presence_handler(instance, stream, func, type_, from_):
    return aioxmpp.stream.presence_handler(stream, type_, from_, func)


def _apply_inbound_message_filter(instance, stream, func):
    return aioxmpp.stream.stanza_filter(
        stream.service_inbound_message_filter,
        func,
        type(instance),
    )


def _apply_inbound_presence_filter(instance, stream, func):
    return aioxmpp.stream.stanza_filter(
        stream.service_inbound_presence_filter,
        func,
        type(instance),
    )


def _apply_outbound_message_filter(instance, stream, func):
    return aioxmpp.stream.stanza_filter(
        stream.service_outbound_message_filter,
        func,
        type(instance),
    )


def _apply_outbound_presence_filter(instance, stream, func):
    return aioxmpp.stream.stanza_filter(
        stream.service_outbound_presence_filter,
        func,
        type(instance),
    )


def _apply_connect_depsignal(instance, stream, func, dependency, signal_name,
                             mode):
    if dependency is aioxmpp.stream.StanzaStream:
        dependency = instance.client.stream
    elif dependency is aioxmpp.node.Client:
        dependency = instance.client
    else:
        dependency = instance.dependencies[dependency]
    signal = getattr(dependency, signal_name)
    if mode is None:
        return signal.context_connect(func)
    else:
        try:
            mode_func, args = mode
        except TypeError:
            pass
        else:
            mode = mode_func(*args)
        return signal.context_connect(func, mode)


def _apply_connect_depfilter(instance, stream, func, dependency, filter_name):
    if dependency is aioxmpp.stream.StanzaStream:
        dependency = instance.client.stream
    else:
        dependency = instance.dependencies[dependency]
    filter_ = getattr(dependency, filter_name)
    return filter_.context_register(func, type(instance))


def _apply_connect_attrsignal(instance, stream, func, descriptor, signal_name,
                              mode):
    obj = descriptor.__get__(instance, type(instance))
    signal = getattr(obj, signal_name)
    if mode is None:
        return signal.context_connect(func)
    else:
        try:
            mode_func, args = mode
        except TypeError:
            pass
        else:
            mode = mode_func(*args)
        return signal.context_connect(func, mode)


def iq_handler(type_, payload_cls):
    """
    Register the decorated function or coroutine function as IQ request
    handler.

    :param type_: IQ type to listen for
    :type type_: :class:`~.IQType`
    :param payload_cls: Payload XSO class to listen for
    :type payload_cls: :class:`~.XSO` subclass
    :raises ValueError: if `payload_cls` is not a registered IQ payload

    If the decorated function is not a coroutine function, it must return an
    awaitable instead.

    .. seealso::

        :meth:`~.StanzaStream.register_iq_request_handler`
            for more details on the `type_` and `payload_cls` arguments, as well
            as behaviour expected from the decorated function.
        :meth:`aioxmpp.IQ.as_payload_class`
            for a way to register a XSO as IQ payload

    .. versionchanged:: 0.10

        The decorator now checks if `payload_cls` is a valid, registered IQ
        payload and raises :class:`ValueError` if not.

    """

    if (not hasattr(payload_cls, "TAG") or
            (aioxmpp.IQ.CHILD_MAP.get(payload_cls.TAG) is not
             aioxmpp.IQ.payload.xq_descriptor) or
            payload_cls not in aioxmpp.IQ.payload._classes):
        raise ValueError(
            "{!r} is not a valid IQ payload "
            "(use IQ.as_payload_class decorator)".format(
                payload_cls,
            )
        )

    def decorator(f):
        add_handler_spec(
            f,
            HandlerSpec(
                (_apply_iq_handler, (type_, payload_cls)),
                require_deps=()
            )
        )
        return f
    return decorator


def message_handler(type_, from_):
    """
    Deprecated alias of :func:`.dispatcher.message_handler`.

    .. deprecated:: 0.9
    """
    import aioxmpp.dispatcher
    return aioxmpp.dispatcher.message_handler(type_, from_)


def presence_handler(type_, from_):
    """
    Deprecated alias of :func:`.dispatcher.presence_handler`.

    .. deprecated:: 0.9
    """
    import aioxmpp.dispatcher
    return aioxmpp.dispatcher.presence_handler(type_, from_)


def inbound_message_filter(f):
    """
    Register the decorated function as a service-level inbound message filter.

    :raise TypeError: if the decorated object is a coroutine function

    .. seealso::

       :class:`StanzaStream`
          for important remarks regarding the use of stanza filters.

    """

    if asyncio.iscoroutinefunction(f):
        raise TypeError(
            "inbound_message_filter must not be a coroutine function"
        )

    add_handler_spec(
        f,
        HandlerSpec(
            (_apply_inbound_message_filter, ())
        ),
    )
    return f


def inbound_presence_filter(f):
    """
    Register the decorated function as a service-level inbound presence filter.

    :raise TypeError: if the decorated object is a coroutine function

    .. seealso::

       :class:`StanzaStream`
          for important remarks regarding the use of stanza filters.

    """

    if asyncio.iscoroutinefunction(f):
        raise TypeError(
            "inbound_presence_filter must not be a coroutine function"
        )

    add_handler_spec(
        f,
        HandlerSpec(
            (_apply_inbound_presence_filter, ())
        ),
    )
    return f


def outbound_message_filter(f):
    """
    Register the decorated function as a service-level outbound message filter.

    :raise TypeError: if the decorated object is a coroutine function

    .. seealso::

       :class:`StanzaStream`
          for important remarks regarding the use of stanza filters.

    """

    if asyncio.iscoroutinefunction(f):
        raise TypeError(
            "outbound_message_filter must not be a coroutine function"
        )

    add_handler_spec(
        f,
        HandlerSpec(
            (_apply_outbound_message_filter, ())
        ),
    )
    return f


def outbound_presence_filter(f):
    """
    Register the decorated function as a service-level outbound presence
    filter.

    :raise TypeError: if the decorated object is a coroutine function

    .. seealso::

       :class:`StanzaStream`
          for important remarks regarding the use of stanza filters.

    """

    if asyncio.iscoroutinefunction(f):
        raise TypeError(
            "outbound_presence_filter must not be a coroutine function"
        )

    add_handler_spec(
        f,
        HandlerSpec(
            (_apply_outbound_presence_filter, ())
        ),
    )
    return f


def _signal_connect_mode(signal, f, defer):
    if isinstance(signal, aioxmpp.callbacks.SyncSignal):
        if not asyncio.iscoroutinefunction(f):
            raise TypeError(
                "a coroutine function is required for this signal"
            )
        if defer:
            raise ValueError(
                "cannot use defer with this signal"
            )
        mode = None
    else:
        if asyncio.iscoroutinefunction(f):
            if defer:
                mode = aioxmpp.callbacks.AdHocSignal.SPAWN_WITH_LOOP, (None,)
            else:
                raise TypeError(
                    "cannot use coroutine function with this signal"
                    " without defer"
                )
        elif defer:
            mode = aioxmpp.callbacks.AdHocSignal.ASYNC_WITH_LOOP, (None,)
        else:
            mode = aioxmpp.callbacks.AdHocSignal.STRONG

    return mode


def _depsignal_spec(class_, signal_name, f, defer):
    signal = getattr(class_, signal_name)

    mode = _signal_connect_mode(signal, f, defer)

    if (class_ is not aioxmpp.stream.StanzaStream and
            class_ is not aioxmpp.node.Client):
        deps = (class_,)
    else:
        deps = ()

    return HandlerSpec(
        (
            _apply_connect_depsignal,
            (
                class_,
                signal_name,
                mode,
            )
        ),
        require_deps=deps,
    )


def depsignal(class_, signal_name, *, defer=False):
    """
    Connect the decorated method or coroutine method to the addressed signal on
    a class on which the service depends.

    :param class_: A service class which is listed in the
                   :attr:`~.Meta.ORDER_AFTER` relationship.
    :type class_: :class:`Service` class or one of the special cases below
    :param signal_name: Attribute name of the signal to connect to
    :type signal_name: :class:`str`
    :param defer: Flag indicating whether deferred execution of the decorated
                  method is desired; see below for details.
    :type defer: :class:`bool`

    The signal is discovered by accessing the attribute with the name
    `signal_name` on the given `class_`. In addition, the following arguments
    are supported for `class_`:

    1. :class:`aioxmpp.stream.StanzaStream`: the corresponding signal of the
       stream of the client running the service is used.

    2. :class:`aioxmpp.Client`: the corresponding signal of the client running
       the service is used.

    If the signal is a :class:`.callbacks.Signal` and `defer` is false, the
    decorated object is connected using the default
    :attr:`~.callbacks.AdHocSignal.STRONG` mode.

    If the signal is a :class:`.callbacks.Signal` and `defer` is true and the
    decorated object is a coroutine function, the
    :attr:`~.callbacks.AdHocSignal.SPAWN_WITH_LOOP` mode with the default
    asyncio event loop is used. If the decorated object is not a coroutine
    function, :attr:`~.callbacks.AdHocSignal.ASYNC_WITH_LOOP` is used instead.

    If the signal is a :class:`.callbacks.SyncSignal`, `defer` must be false
    and the decorated object must be a coroutine function.

    .. versionchanged:: 0.9

       Support for :class:`aioxmpp.stream.StanzaStream` and
       :class:`aioxmpp.Client` as `class_` argument was added.
    """

    def decorator(f):
        add_handler_spec(
            f,
            _depsignal_spec(class_, signal_name, f, defer)
        )
        return f
    return decorator


def _attrsignal_spec(descriptor, signal_name, f, defer):
    signal = getattr(descriptor.value_type, signal_name)
    mode = _signal_connect_mode(signal, f, defer)

    return HandlerSpec(
        (
            _apply_connect_attrsignal,
            (
                descriptor,
                signal_name,
                mode
            )
        ),
        is_unique=True,
        require_deps=(),
    )


def attrsignal(descriptor, signal_name, *, defer=False):
    """
    Connect the decorated method or coroutine method to the addressed signal on
    a descriptor.

    :param descriptor: The descriptor to connect to.
    :type descriptor: :class:`Descriptor` subclass.
    :param signal_name: Attribute name of the signal to connect to
    :type signal_name: :class:`str`
    :param defer: Flag indicating whether deferred execution of the decorated
                  method is desired; see below for details.
    :type defer: :class:`bool`

    The signal is discovered by accessing the attribute with the name
    `signal_name` on the :attr:`~Descriptor.value_type` of the `descriptor`.

    During instantiation of the service, the value of the descriptor is used
    to obtain the signal and then the decorated method is connected to the
    signal.

    If the signal is a :class:`.callbacks.Signal` and `defer` is false, the
    decorated object is connected using the default
    :attr:`~.callbacks.AdHocSignal.STRONG` mode.

    If the signal is a :class:`.callbacks.Signal` and `defer` is true and the
    decorated object is a coroutine function, the
    :attr:`~.callbacks.AdHocSignal.SPAWN_WITH_LOOP` mode with the default
    asyncio event loop is used. If the decorated object is not a coroutine
    function, :attr:`~.callbacks.AdHocSignal.ASYNC_WITH_LOOP` is used instead.

    If the signal is a :class:`.callbacks.SyncSignal`, `defer` must be false
    and the decorated object must be a coroutine function.

    .. versionadded:: 0.9
    """
    def decorator(f):
        add_handler_spec(
            f,
            _attrsignal_spec(descriptor, signal_name, f, defer)
        )
        return f
    return decorator


def _depfilter_spec(class_, filter_name):
    require_deps = ()
    if class_ is not aioxmpp.stream.StanzaStream:
        require_deps = (class_,)

    return HandlerSpec(
        (
            _apply_connect_depfilter,
            (
                class_,
                filter_name,
            )
        ),
        is_unique=True,
        require_deps=require_deps,
    )


def depfilter(class_, filter_name):
    """
    Register the decorated method at the addressed :class:`~.callbacks.Filter`
    on a class on which the service depends.

    :param class_: A service class which is listed in the
                   :attr:`~.Meta.ORDER_AFTER` relationship.
    :type class_: :class:`Service` class or
                  :class:`aioxmpp.stream.StanzaStream`
    :param filter_name: Attribute name of the filter to register at
    :type filter_name: :class:`str`

    The filter at which the decorated method is registered is discovered by
    accessing the attribute with the name `filter_name` on the instance of the
    dependent class `class_`. If `class_` is
    :class:`aioxmpp.stream.StanzaStream`, the filter is searched for on the
    stream (and no dependendency needs to be declared).

    .. versionadded:: 0.9
    """
    spec = _depfilter_spec(class_, filter_name)

    def decorator(f):
        add_handler_spec(
            f,
            spec,
        )
        return f

    return decorator


def is_iq_handler(type_, payload_cls, coro):
    """
    Return true if `coro` has been decorated with :func:`iq_handler` for the
    given `type_` and `payload_cls`.
    """

    try:
        handlers = get_magic_attr(coro)
    except AttributeError:
        return False

    return HandlerSpec(
        (_apply_iq_handler, (type_, payload_cls)),
    ) in handlers


def is_message_handler(type_, from_, cb):
    """
    Deprecated alias of :func:`.dispatcher.is_message_handler`.

    .. deprecated:: 0.9
    """
    import aioxmpp.dispatcher
    return aioxmpp.dispatcher.is_message_handler(type_, from_, cb)


def is_presence_handler(type_, from_, cb):
    """
    Deprecated alias of :func:`.dispatcher.is_presence_handler`.

    .. deprecated:: 0.9
    """
    import aioxmpp.dispatcher
    return aioxmpp.dispatcher.is_presence_handler(type_, from_, cb)


def is_inbound_message_filter(cb):
    """
    Return true if `cb` has been decorated with :func:`inbound_message_filter`.
    """

    try:
        handlers = get_magic_attr(cb)
    except AttributeError:
        return False

    return HandlerSpec(
        (_apply_inbound_message_filter, ())
    ) in handlers


def is_inbound_presence_filter(cb):
    """
    Return true if `cb` has been decorated with
    :func:`inbound_presence_filter`.
    """

    try:
        handlers = get_magic_attr(cb)
    except AttributeError:
        return False

    return HandlerSpec(
        (_apply_inbound_presence_filter, ())
    ) in handlers


def is_outbound_message_filter(cb):
    """
    Return true if `cb` has been decorated with
    :func:`outbound_message_filter`.
    """

    try:
        handlers = get_magic_attr(cb)
    except AttributeError:
        return False

    return HandlerSpec(
        (_apply_outbound_message_filter, ())
    ) in handlers


def is_outbound_presence_filter(cb):
    """
    Return true if `cb` has been decorated with
    :func:`outbound_presence_filter`.
    """

    try:
        handlers = get_magic_attr(cb)
    except AttributeError:
        return False

    return HandlerSpec(
        (_apply_outbound_presence_filter, ())
    ) in handlers


def is_depsignal_handler(class_, signal_name, cb, *, defer=False):
    """
    Return true if `cb` has been decorated with :func:`depsignal` for the given
    signal, class and connection mode.
    """
    try:
        handlers = get_magic_attr(cb)
    except AttributeError:
        return False

    return _depsignal_spec(class_, signal_name, cb, defer) in handlers


def is_depfilter_handler(class_, filter_name, filter_):
    """
    Return true if `filter_` has been decorated with :func:`depfilter` for the
    given filter and class.
    """
    try:
        handlers = get_magic_attr(filter_)
    except AttributeError:
        return False

    return _depfilter_spec(class_, filter_name) in handlers


def is_attrsignal_handler(descriptor, signal_name, cb, *, defer=False):
    """
    Return true if `cb` has been decorated with :func:`attrsignal` for the
    given signal, descriptor and connection mode.
    """
    try:
        handlers = get_magic_attr(cb)
    except AttributeError:
        return False

    return _attrsignal_spec(descriptor, signal_name, cb, defer) in handlers



########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


namespaces.xep0049 = "jabber:iq:private"


@aioxmpp.IQ.as_payload_class
class Query(xso.XSO):
    """
    The XSO for queries to private XML storage.

    .. automethod:: as_payload_class
    """
    TAG = (namespaces.xep0049, "query")

    registered_payload = xso.Child([], strict=True)
    unregistered_payload = xso.Collector()

    def __init__(self, payload):
        self.registered_payload = payload

    @classmethod
    def as_payload_class(mycls, xso_class):
        """
        Register the given class `xso_class` as possible payload
        for private XML storage.

        Return `xso_class`, to allow this to be used as a decorator.
        """
        mycls.register_child(
            Query.registered_payload,
            xso_class
        )

        return xso_class



########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp

import aioxmpp.im.service as im_service

from aioxmpp.testutils import (
    make_connected_client,
    make_listener,
)


class TestConversationService(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.s = im_service.ConversationService(self.cc)
        self.listener = make_listener(self.s)

    def tearDown(self):
        del self.s
        del self.cc

    def test_init(self):
        self.assertSequenceEqual(
            list(self.s.conversations),
            [],
        )

    def test__add_conversation_emits_on_conversation_added(self):
        conv = unittest.mock.Mock()
        self.s._add_conversation(conv)
        self.listener.on_conversation_added.assert_called_once_with(
            conv,
        )

    def test__add_conversation_adds_conversation_to_conversations(self):
        conv = unittest.mock.Mock()
        self.s._add_conversation(conv)

        self.assertCountEqual(
            self.s.conversations,
            [
                conv,
            ]
        )

    def test__add_conversation_makes_it_accessible_by_jid(self):
        conv = unittest.mock.Mock()
        conv.jid = unittest.mock.sentinel.jid
        self.s._add_conversation(conv)

        self.assertIs(self.s.get_conversation(conv.jid), conv)

    def test_conversation_not_accessible_after_exit(self):
        conv = unittest.mock.Mock()
        conv.jid = unittest.mock.sentinel.jid
        self.s._add_conversation(conv)

        _, (cb, ), _ = conv.on_exit.mock_calls[-1]
        cb()

        with self.assertRaises(KeyError):
            self.s.get_conversation(conv.jid)

    def test_forwards_message_events(self):
        conv = unittest.mock.Mock(["on_failure", "on_exit", "jid",
                                   "on_message"])
        conv.jid = unittest.mock.sentinel.jid

        self.s._add_conversation(conv)

        conv.on_message.connect.assert_called_once_with(
            unittest.mock.ANY,
        )

        _, (cb, ), _ = conv.on_message.mock_calls[-1]

        cb(unittest.mock.sentinel.message,
           unittest.mock.sentinel.member,
           unittest.mock.sentinel.source,
           tracker=unittest.mock.sentinel.tracker,
           foo_bar=unittest.mock.sentinel.foobar)

        self.listener.on_message.assert_called_once_with(
            conv,
            unittest.mock.sentinel.message,
            unittest.mock.sentinel.member,
            unittest.mock.sentinel.source,
            tracker=unittest.mock.sentinel.tracker,
            foo_bar=unittest.mock.sentinel.foobar
        )

    def test_on_conversation_added_sees_added_conversation(self):
        captured_conversations = None

        def cb(conv):
            nonlocal captured_conversations
            captured_conversations = list(self.s.conversations)

        conv = unittest.mock.Mock()
        self.s.on_conversation_added.connect(cb)
        self.s._add_conversation(conv)

        self.assertIn(
            conv,
            captured_conversations,
        )

    def test_removes_added_conversation_on_exit(self):
        conv = unittest.mock.Mock()
        self.s._add_conversation(conv)

        conv.on_exit.connect.assert_called_once_with(
            unittest.mock.ANY
        )

        _, (cb, ), _ = conv.on_exit.mock_calls[-1]

        # should ignore its arguments
        cb(unittest.mock.sentinel.foo, bar=unittest.mock.sentinel.fnord)

        self.assertCountEqual(
            self.s.conversations,
            [
            ]
        )

    def test_disconnects_handlers_on_exit(self):
        conv = unittest.mock.Mock(["on_exit", "on_failure", "on_message",
                                   "jid"])
        self.s._add_conversation(conv)

        conv.on_exit.connect.assert_called_once_with(
            unittest.mock.ANY
        )

        _, (cb, ), _ = conv.on_exit.mock_calls[-1]

        cb()

        conv.on_exit.disconnect.assert_called_once_with(
            conv.on_exit.connect(),
        )

        conv.on_failure.disconnect.assert_called_once_with(
            conv.on_failure.connect(),
        )

        conv.on_message.disconnect.assert_called_once_with(
            conv.on_message.connect(),
        )

    def test_disconnects_handlers_on_failure(self):
        conv = unittest.mock.Mock(["on_exit", "on_failure", "on_message",
                                   "jid"])
        self.s._add_conversation(conv)

        conv.on_failure.connect.assert_called_once_with(
            unittest.mock.ANY
        )

        _, (cb, ), _ = conv.on_failure.mock_calls[-1]

        cb()

        conv.on_exit.disconnect.assert_called_once_with(
            conv.on_exit.connect(),
        )

        conv.on_failure.disconnect.assert_called_once_with(
            conv.on_failure.connect(),
        )

        conv.on_message.disconnect.assert_called_once_with(
            conv.on_message.connect(),
        )

    def test_removes_added_conversation_on_failure(self):
        conv = unittest.mock.Mock()
        self.s._add_conversation(conv)

        conv.on_failure.connect.assert_called_once_with(
            unittest.mock.ANY
        )

        _, (cb, ), _ = conv.on_failure.mock_calls[-1]

        # should ignore its arguments
        cb(unittest.mock.sentinel.foo, unittest.mock.sentinel.bar,
           bar=unittest.mock.sentinel.fnord)

        self.assertCountEqual(
            self.s.conversations,
            [
            ]
        )



########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio

import aioxmpp.service

from aioxmpp.utils import namespaces


class SHIMService(aioxmpp.service.Service):
    """
    This service implements :xep:`131` feature advertisment.

    It registers the ``http://jabber.org/protocol/shim`` node with the
    :class:`.DiscoServer`. It publishes the supported headers on that node as
    specified in the XEP.

    To announce supported headers, use the :meth:`register_header` and
    :meth:`unregister_header` methods.

    .. automethod:: register_header

    .. automethod:: unregister_header

    .. versionchanged:: 0.8

       This class was formerly known as :class:`aioxmpp.shim.Service`. It
       is still available under that name, but the alias will be removed in
       1.0.
    """

    ORDER_AFTER = [aioxmpp.DiscoServer]

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)

        self._disco = self.dependencies[aioxmpp.DiscoServer]
        self._disco.register_feature(namespaces.xep0131_shim)

        self._node = aioxmpp.disco.StaticNode()
        self._disco.mount_node(
            namespaces.xep0131_shim,
            self._node
        )

    @asyncio.coroutine
    def _shutdown(self):
        self._disco.unregister_feature(namespaces.xep0131_shim)
        self._disco.unmount_node(namespaces.xep0131_shim)
        yield from super()._shutdown()

    def register_header(self, name):
        """
        Register support for the SHIM header with the given `name`.

        If the header has already been registered as supported,
        :class:`ValueError` is raised.
        """

        self._node.register_feature(
            "#".join([namespaces.xep0131_shim, name])
        )

    def unregister_header(self, name):
        """
        Unregister support for the SHIM header with the given `name`.

        If the header is currently not registered as supported,
        :class:`KeyError` is raised.
        """

        self._node.unregister_feature(
            "#".join([namespaces.xep0131_shim, name])
        )



########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces

from ..misc import Forwarded
from ..stanza import Message, IQ


namespaces.xep0280_carbons_2 = "urn:xmpp:carbons:2"


@IQ.as_payload_class
class Enable(xso.XSO):
    TAG = (namespaces.xep0280_carbons_2, "enable")


@IQ.as_payload_class
class Disable(xso.XSO):
    TAG = (namespaces.xep0280_carbons_2, "disable")


class _CarbonsWrapper(xso.XSO):
    forwarded = xso.Child([Forwarded])

    @property
    def stanza(self):
        """
        The wrapped stanza, usually a :class:`aioxmpp.Message`.

        Internally, this accesses the :attr:`~.misc.Forwarded.stanza` attribute
        of :attr:`forwarded`. If :attr:`forwarded` is :data:`None`, reading
        this attribute returns :data:`None`. Writing to this attribute creates
        a new :class:`~.misc.Forwarded` object if necessary, but re-uses an
        existing object if available.
        """
        if self.forwarded is None:
            return None
        return self.forwarded.stanza

    @stanza.setter
    def stanza(self, value):
        if self.forwarded is None:
            self.forwarded = Forwarded()
        self.forwarded.stanza = value


class Sent(_CarbonsWrapper):
    """
    Wrap a stanza which was sent by another entity of the same account.

    :class:`Sent` XSOs are available in Carbon messages at
    :attr:`aioxmpp.Message.xep0280_sent`.

    .. autoattribute:: stanza

    .. attribute:: forwarded

       The full :class:`~.misc.Forwarded` object which holds the sent stanza.

    """

    TAG = (namespaces.xep0280_carbons_2, "sent")


class Received(_CarbonsWrapper):
    """
    Wrap a stanza which was received by another entity of the same account.

    :class:`Received` XSOs are available in Carbon messages at
    :attr:`aioxmpp.Message.xep0280_received`.

    .. autoattribute:: stanza

    .. attribute:: forwarded

       The full :class:`~.misc.Forwarded` object which holds the received
       stanza.

    """
    TAG = (namespaces.xep0280_carbons_2, "received")


Message.xep0280_sent = xso.Child([Sent])
Message.xep0280_received = xso.Child([Received])



########################################################################
# File name: retrieve_avatar.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import configparser

import aioxmpp
import aioxmpp.avatar

from framework import Example, exec_example


class Avatar(Example):
    def prepare_argparse(self):
        super().prepare_argparse()

        # this gives a nicer name in argparse errors
        def jid(s):
            return aioxmpp.JID.fromstr(s)

        self.argparse.add_argument(
            "output_file",
            help="the file the retrieved avatar image will be written to."
        )

        self.argparse.add_argument(
            "--remote-jid",
            type=jid,
            help="the jid of which to retrieve the avatar"
        )

    def configure(self):
        super().configure()

        self.output_file = self.args.output_file
        self.remote_jid = self.args.remote_jid
        if self.remote_jid is None:
            try:
                self.remote_jid = aioxmpp.JID.fromstr(
                    self.config.get("avatar", "remote_jid")
                )
            except (configparser.NoSectionError,
                    configparser.NoOptionError):
                self.remote_jid = aioxmpp.JID.fromstr(
                    input("Remote JID> ")
                )

    def make_simple_client(self):
        client = super().make_simple_client()
        self.avatar = client.summon(aioxmpp.avatar.AvatarService)
        return client

    @asyncio.coroutine
    def run_simple_example(self):
        metadata = yield from self.avatar.get_avatar_metadata(
            self.remote_jid
        )

        for metadatum in metadata:
            if metadatum.can_get_image_bytes_via_xmpp:
                image = yield from metadatum.get_image_bytes()
                with open(self.output_file, "wb") as avatar_image:
                    avatar_image.write(image)
                return

        print("retrieving avatar failed: no avatar available via xmpp")

    @asyncio.coroutine
    def run_example(self):
        yield from super().run_example()


if __name__ == "__main__":
    exec_example(Avatar())



########################################################################
# File name: ibr_test.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import getpass

try:
    import readline
except ImportError:
    pass

import aioxmpp
import aioxmpp.ibr as ibr


async def get_fields(jid):
    metadata = aioxmpp.make_security_layer(None)
    _, stream, features = await aioxmpp.node.connect_xmlstream(jid, metadata)
    reply = await ibr.get_registration_fields(stream)
    print(ibr.get_used_fields(reply))


async def get_info(jid, password):
    client = aioxmpp.PresenceManagedClient(
        jid,
        aioxmpp.make_security_layer(password)
    )

    async with client.connected() as stream:
        service = ibr.RegistrationService(stream)
        reply = await service.get_client_info()
        print("Username: " + reply.username)


async def register(jid, password):
    metadata = aioxmpp.make_security_layer(None)
    _, stream, features = await aioxmpp.node.connect_xmlstream(jid, metadata)

    query = ibr.Query(jid.localpart, password)
    await ibr.register(stream, query)
    print("Registered")


async def change_password(jid, old_password, new_password):
    client = aioxmpp.PresenceManagedClient(
        jid,
        aioxmpp.make_security_layer(old_password)
    )

    async with client.connected() as stream:
        service = ibr.RegistrationService(stream)
        await service.change_pass(new_password)

    print("Password changed")


async def cancel(jid, password):
    client = aioxmpp.PresenceManagedClient(
        jid,
        aioxmpp.make_security_layer(password)
    )

    async with client.connected() as stream:
        service = ibr.RegistrationService(stream)
        await service.cancel_registration()


if __name__ == "__main__":
    local_jid = aioxmpp.JID.fromstr(input("local JID> "))
    password = getpass.getpass()

    loop = asyncio.get_event_loop()

    try:
        loop.run_until_complete(cancel(local_jid, password))
    except:
        pass

    loop.run_until_complete(get_fields(local_jid, ))

    loop.run_until_complete(register(local_jid, password))

    loop.run_until_complete(get_info(local_jid, password))

    loop.run_until_complete(change_password(local_jid, password, "aaa"))

    try:
        loop.run_until_complete(cancel(local_jid, "aaa"))
    except:
        pass



########################################################################
# File name: service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import logging

import aioxmpp.service

import aioxmpp.callbacks as callbacks
import aioxmpp.errors as errors
import aioxmpp.stanza as stanza
import aioxmpp.structs as structs

from aioxmpp.utils import namespaces

from . import xso as roster_xso


logger = logging.getLogger(__name__)


_Sentinel = object()


class Item:
    """
    Represent an entry in the roster. These entries are mutable, see the
    documentation of :class:`Service` for details on the lifetime of
    :class:`Item` instances within a :class:`Service` instance.

    .. attribute:: jid

       The :class:`~aioxmpp.JID` of the entry. This is always a bare
       JID.

    .. attribute:: name

       The display name of the entry, if any.

    .. attribute:: groups

       A :class:`set` of names of groups in which the roster entry is.

    .. attribute:: subscription

       The subscription status of the entry. One of ``"none"``, ``"to"``,
       ``"from"`` and ``"both"`` (in contrast to :class:`.xso.Item`,
       ``"remove"`` cannot occur here).

    .. attribute:: ask

       The ``ask`` attribute of the roster entry.

    .. attribute:: approved

       The ``approved`` attribute of the roster entry.

    The data of a roster entry can conveniently be exported to JSON:

    .. automethod:: export_as_json

    To mutate the roster entry, some handy methods are provided:

    .. automethod:: update_from_json

    .. automethod:: update_from_xso_item

    To create a roster entry from a :class:`.xso.Item`, use the
    :meth:`from_xso_item` class method.

    .. automethod:: from_xso_item

    .. note::

       Do not confuse this with the XSO :class:`.xso.Item`.

    """
    def __init__(self, jid, *,
                 approved=False,
                 ask=None,
                 subscription="none",
                 name=None,
                 groups=()):
        super().__init__()
        self.jid = jid
        self.subscription = subscription
        self.approved = approved
        self.ask = ask
        self.name = name
        self.groups = set(groups)

    def update_from_xso_item(self, xso_item):
        """
        Update the attributes (except :attr:`jid`) with the values obtained
        from the gixen `xso_item`.

        `xso_item` must be a valid :class:`.xso.Item` instance.
        """
        self.subscription = xso_item.subscription
        self.approved = xso_item.approved
        self.ask = xso_item.ask
        self.name = xso_item.name
        self.groups = {group.name for group in xso_item.groups}

    @classmethod
    def from_xso_item(cls, xso_item):
        """
        Create a :class:`Item` with the :attr:`jid` set to the
        :attr:`.xso.Item.jid` obtained from `xso_item`. Then update that
        instance with `xso_item` using :meth:`update_from_xso_item` and return
        it.
        """
        item = cls(xso_item.jid)
        item.update_from_xso_item(xso_item)
        return item

    def export_as_json(self):
        """
        Return a :mod:`json`-compatible dictionary which contains the
        attributes of this :class:`Item` except its JID.
        """
        result = {
            "subscription": self.subscription,
        }

        if self.name:
            result["name"] = self.name

        if self.ask is not None:
            result["ask"] = self.ask

        if self.approved:
            result["approved"] = self.approved

        if self.groups:
            result["groups"] = sorted(self.groups)

        return result

    def update_from_json(self, data):
        """
        Update the attributes of this :class:`Item` using the values obtained
        from the dictionary `data`.

        The format of `data` should be the same as the format returned by
        :meth:`export_as_json`.
        """
        self.subscription = data.get("subscription", "none")
        self.approved = bool(data.get("approved", False))
        self.ask = data.get("ask", None)
        self.name = data.get("name", None)
        self.groups = set(data.get("groups", []))


class RosterClient(aioxmpp.service.Service):
    """
    A roster client :class:`aioxmpp.service.Service`.

    The interaction with a roster service happens mainly by accessing the
    attributes holding the state and using the events to be notified of state
    changes:

    Attributes for accessing the roster:

    .. attribute:: items

       A dictionary mapping :class:`~aioxmpp.JID` instances to corresponding
       :class:`Item` instances.

    .. attribute:: groups

       A dictionary which allows group-based access to :class:`Item`
       instances. The dictionaries keys are the names of the groups, the values
       are :class:`set` instances, which hold the :class:`Item` instances in
       that group.

       At no point one can observe empty :class:`set` instances in this
       dictionary.

    The :class:`Item` instances stay the same, as long as they represent the
    identical roster entry on the remote side. That is, if the name or
    subscription state are changed in the server side roster, the :class:`Item`
    instance stays the same, but the attributes are mutated. However, if the
    entry is removed from the server roster and re-added later for the same
    JID, it will be a different :class:`Item` instance.

    Signals:

    .. signal:: on_initial_roster_received()

       Fires when the initial roster has been received. Note that if roster
       versioning is used, the initial roster may not be up-to-date. The server
       is allowed to tell the client to re-use its local state and deliver
       changes using roster pushes. In that case, the
       :meth:`on_initial_roster_received` event fires immediately, so that the
       user sees whatever roster has been set up for versioning before the
       stream was established; updates pushed by the server are delivered using
       the normal events.

       The roster data has already been imported at the time the callback is
       fired.

       Note that the initial roster is diffed against whatever is in the local
       store and events are fired just like for normal push updates. Thus, in
       general, you won’t need this signal; it might be better to listen for
       the events below.

    .. signal:: on_entry_added(item)

       Fires when an `item` has been added to the roster. The attributes of the
       `item` are up-to-date when this callback fires.

       When the event fires, the bookkeeping structures are already updated.
       This implies that :meth:`on_group_added` is called before
       :meth:`on_entry_added` if the entry adds a new group.

    .. signal:: on_entry_name_changed(item)

       Fires when a roster update changed the name of the `item`. The new name
       is already applied to the `item`.

    .. signal:: on_entry_subscription_state_changed(item)

       Fires when a roster update changes any of the :attr:`Item.subscription`,
       :attr:`Item.ask` or :attr:`Item.approved` attributes. The new values are
       already applied to `item`.

       The event always fires once per update, even if the update changes
       more than one of the above attributes.

    .. signal:: on_entry_added_to_group(item, group_name)

       Fires when an update adds an `item` to a group. The :attr:`Item.groups`
       attribute is already updated (not only with this, but also other group
       updates, including removals) when this event is fired.

       The event fires for each added group in an update, thus it may fire more
       than once per update.

       The name of the new group is in `group_name`.

       At the time the event fires, the bookkeeping structures for the group
       are already updated; this implies that :meth:`on_group_added` fires
       *before* :meth:`on_entry_added_to_group` if the entry added a new group.

    .. signal:: on_entry_removed_from_group(item, group_name)

       Fires when an update removes an `item` from a group. The
       :attr:`Item.groups` attribute is already updated (not only with this,
       but also other group updates, including additions) when this event is
       fired.

       The event fires for each removed group in an update, thus it may fire
       more than once per update.

       The name of the new group is in `group_name`.

       At the time the event fires, the bookkeeping structures are already
       updated; this implies that :meth:`on_group_removed` fires *before*
       :meth:`on_entry_removed_from_group` if the removal of an entry from a
       group causes the group to vanish.

    .. signal:: on_entry_removed(item)

       Fires after an entry has been removed from the roster. The entry is
       already removed from all bookkeeping structures, but the values on the
       `item` object are the same as right before the removal.

       This implies that :meth:`on_group_removed` fires *before*
       :meth:`on_entry_removed` if the removal of an entry causes a group to
       vanish.

    .. signal:: on_group_added(group)

        Fires after a new group has been added to the bookkeeping structures.

        :param group: Name of the new group.
        :type group: :class:`str`

        At the time the event fires, the group is empty.

        .. versionadded:: 0.9

    .. signal:: on_group_removed(group)

        Fires after a new group has been removed from the bookkeeping
        structures.

        :param group: Name of the old group.
        :type group: :class:`str`

        At the time the event fires, the group is empty.

        .. versionadded:: 0.9

    Modifying roster contents:

    .. automethod:: set_entry

    .. automethod:: remove_entry

    Managing presence subscriptions:

    .. automethod:: approve

    .. automethod:: subscribe

    .. signal:: on_subscribe(stanza)

       Fires when a peer requested a subscription. The whole stanza recevied is
       included as `stanza`.

       .. seealso::

          To approve a subscription request, use :meth:`approve`.

    .. signal:: on_subscribed(stanza)

       Fires when a peer has confirmed a previous subscription request. The
       ``"subscribed"`` stanza is included as `stanza`.

    .. signal:: on_unsubscribe(stanza)

       Fires when a peer cancelled their subscription for our presence. As per
       :rfc:`6121`, the server forwards the ``"unsubscribe"`` presence stanza
       (which is included as `stanza` argument) *before* sending the roster
       push.

       Unless your application is interested in the specific cause of a
       subscription state change, it is not neccessary to use this signal; the
       subscription state change will be covered by
       :meth:`on_entry_subscription_state_changed`.

    .. signal:: on_unsubscribed(stanza)

       Fires when a peer cancelled our subscription. As per :rfc:`6121`, the
       server forwards the ``"unsubscribed"`` presence stanza (which is
       included as `stanza` argument) *before* sending the roster push.

       Unless your application is interested in the specific cause of a
       subscription state change, it is not neccessary to use this signal; the
       subscription state change will be covered by
       :meth:`on_entry_subscription_state_changed`.

    Import/Export of roster data:

    .. automethod:: export_as_json

    .. automethod:: import_from_json

    To make use of roster versioning, use the above two methods. The general
    workflow is to :meth:`export_as_json` the roster after disconnecting and
    storing it for the next connection attempt. **Before** connecting, the
    stored data needs to be loaded using :meth:`import_from_json`. This only
    needs to happen after a new :class:`Service` has been created, as roster
    services won’t delete roster contents between two connections on the same
    :class:`.Client` instance.

    .. versionchanged:: 0.8

       This class was formerly known as :class:`aioxmpp.roster.Service`. It
       is still available under that name, but the alias will be removed in
       1.0.
    """

    ORDER_AFTER = [
        aioxmpp.dispatcher.SimplePresenceDispatcher,
    ]

    on_initial_roster_received = callbacks.Signal()
    on_entry_name_changed = callbacks.Signal()
    on_entry_subscription_state_changed = callbacks.Signal()
    on_entry_removed = callbacks.Signal()
    on_entry_added = callbacks.Signal()
    on_entry_added_to_group = callbacks.Signal()
    on_entry_removed_from_group = callbacks.Signal()

    on_group_added = callbacks.Signal()
    on_group_removed = callbacks.Signal()

    on_subscribed = callbacks.Signal()
    on_subscribe = callbacks.Signal()
    on_unsubscribed = callbacks.Signal()
    on_unsubscribe = callbacks.Signal()

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)

        self._bse_token = client.before_stream_established.connect(
            self._request_initial_roster
        )

        self.__roster_lock = asyncio.Lock()

        self.items = {}
        self.groups = {}
        self.version = None

    def _update_entry(self, xso_item):
        try:
            stored_item = self.items[xso_item.jid]
        except KeyError:
            stored_item = Item.from_xso_item(xso_item)
            self.items[xso_item.jid] = stored_item
            for group in stored_item.groups:
                try:
                    group_members = self.groups[group]
                except KeyError:
                    group_members = self.groups.setdefault(group, set())
                    self.on_group_added(group)
                group_members.add(stored_item)
            self.on_entry_added(stored_item)
            return

        to_call = []

        if stored_item.name != xso_item.name:
            to_call.append(self.on_entry_name_changed)

        if (stored_item.subscription != xso_item.subscription or
                stored_item.approved != xso_item.approved or
                stored_item.ask != xso_item.ask):
            to_call.append(self.on_entry_subscription_state_changed)

        old_groups = set(stored_item.groups)

        stored_item.update_from_xso_item(xso_item)

        new_groups = set(stored_item.groups)

        removed_from_groups = old_groups - new_groups
        added_to_groups = new_groups - old_groups

        for cb in to_call:
            cb(stored_item)

        for group in added_to_groups:
            try:
                group_members = self.groups[group]
            except KeyError:
                group_members = self.groups.setdefault(group, set())
                self.on_group_added(group)
            group_members.add(stored_item)
            self.on_entry_added_to_group(stored_item, group)

        for group in removed_from_groups:
            groupset = self.groups[group]
            groupset.remove(stored_item)
            if not groupset:
                del self.groups[group]
                self.on_group_removed(group)
            self.on_entry_removed_from_group(stored_item, group)

    @aioxmpp.service.iq_handler(
        aioxmpp.structs.IQType.SET,
        roster_xso.Query)
    @asyncio.coroutine
    def handle_roster_push(self, iq):
        if iq.from_ and iq.from_ != self.client.local_jid.bare():
            raise errors.XMPPAuthError(errors.ErrorCondition.FORBIDDEN)

        request = iq.payload

        with (yield from self.__roster_lock):
            for item in request.items:
                if item.subscription == "remove":
                    try:
                        old_item = self.items.pop(item.jid)
                    except KeyError:
                        pass
                    else:
                        self._remove_from_groups(old_item, old_item.groups)
                        self.on_entry_removed(old_item)
                else:
                    self._update_entry(item)

            self.version = request.ver

    @aioxmpp.dispatcher.presence_handler(
        aioxmpp.structs.PresenceType.SUBSCRIBE,
        None)
    def handle_subscribe(self, stanza):
        self.on_subscribe(stanza)

    @aioxmpp.dispatcher.presence_handler(
        aioxmpp.structs.PresenceType.SUBSCRIBED,
        None)
    def handle_subscribed(self, stanza):
        self.on_subscribed(stanza)

    @aioxmpp.dispatcher.presence_handler(
        aioxmpp.structs.PresenceType.UNSUBSCRIBED,
        None)
    def handle_unsubscribed(self, stanza):
        self.on_unsubscribed(stanza)

    @aioxmpp.dispatcher.presence_handler(
        aioxmpp.structs.PresenceType.UNSUBSCRIBE,
        None)
    def handle_unsubscribe(self, stanza):
        self.on_unsubscribe(stanza)

    def _remove_from_groups(self, item_to_remove, groups):
        for group in groups:
            try:
                group_members = self.groups[group]
            except KeyError:
                continue
            group_members.remove(item_to_remove)
            if not group_members:
                del self.groups[group]
                self.on_group_removed(group)

    @asyncio.coroutine
    def _request_initial_roster(self):
        iq = stanza.IQ(type_=structs.IQType.GET)
        iq.payload = roster_xso.Query()

        with (yield from self.__roster_lock):
            logger.debug("requesting initial roster")
            if self.client.stream_features.has_feature(
                    roster_xso.RosterVersioningFeature):
                logger.debug("requesting incremental updates (old ver = %s)",
                             self.version)
                iq.payload.ver = self.version

            response = yield from self.client.send(
                iq,
                timeout=self.client.negotiation_timeout.total_seconds()
            )

            if response is None:
                logger.debug("roster will be updated incrementally")
                self.on_initial_roster_received()
                return True

            self.version = response.ver
            logger.debug("roster update received (new ver = %s)", self.version)

            actual_jids = {item.jid for item in response.items}
            known_jids = set(self.items.keys())

            removed_jids = known_jids - actual_jids
            logger.debug("jids dropped: %r", removed_jids)

            for removed_jid in removed_jids:
                old_item = self.items.pop(removed_jid)
                self._remove_from_groups(old_item, old_item.groups)
                self.on_entry_removed(old_item)

            logger.debug("jids updated: %r", actual_jids - removed_jids)
            for item in response.items:
                self._update_entry(item)

            self.on_initial_roster_received()
            return True

    def export_as_json(self):
        """
        Export the whole roster as currently stored on the client side into a
        JSON-compatible dictionary and return that dictionary.
        """
        return {
            "items": {
                str(jid): item.export_as_json()
                for jid, item in self.items.items()
            },
            "ver": self.version
        }

    def import_from_json(self, data):
        """
        Replace the current roster with the :meth:`export_as_json`-compatible
        dictionary in `data`.

        No events are fired during this activity. After this method completes,
        the whole roster contents are exchanged with the contents from `data`.

        Also, no data is transferred to the server; this method is intended to
        be used for roster versioning. See below (in the docs of
        :class:`Service`).
        """
        self.version = data.get("ver", None)

        self.items.clear()
        self.groups.clear()
        for jid, data in data.get("items", {}).items():
            jid = structs.JID.fromstr(jid)
            item = Item(jid)
            item.update_from_json(data)
            self.items[jid] = item
            for group in item.groups:
                self.groups.setdefault(group, set()).add(item)

    @asyncio.coroutine
    def set_entry(self, jid, *,
                  name=_Sentinel,
                  add_to_groups=frozenset(),
                  remove_from_groups=frozenset(),
                  timeout=None):
        """
        Set properties of a roster entry or add a new roster entry. The roster
        entry is identified by its bare `jid`.

        If an entry already exists, all values default to those stored in the
        existing entry. For example, if no `name` is given, the current name of
        the entry is re-used, if any.

        If the entry does not exist, it will be created on the server side.

        The `remove_from_groups` and `add_to_groups` arguments have to be based
        on the locally cached state, as XMPP does not support sending
        diffs. `remove_from_groups` takes precedence over `add_to_groups`.

        `timeout` is the time in seconds to wait for a confirmation by the
        server.

        Note that the changes may not be visible immediately after his
        coroutine returns in the :attr:`items` and :attr:`groups`
        attributes. The :class:`Service` waits for the "official" roster push
        from the server for updating the data structures and firing events, to
        ensure that consistent state with other clients is achieved.

        This may raise arbitrary :class:`.errors.XMPPError` exceptions if the
        server replies with an error and also any kind of connection error if
        the connection gets fatally terminated while waiting for a response.
        """

        existing = self.items.get(jid, Item(jid))

        post_groups = (existing.groups | add_to_groups) - remove_from_groups
        post_name = existing.name
        if name is not _Sentinel:
            post_name = name

        item = roster_xso.Item(
            jid=jid,
            name=post_name,
            groups=[
                roster_xso.Group(name=group_name)
                for group_name in post_groups
            ])

        yield from self.client.send(
            stanza.IQ(
                structs.IQType.SET,
                payload=roster_xso.Query(items=[
                    item
                ])
            ),
            timeout=timeout
        )

    @asyncio.coroutine
    def remove_entry(self, jid, *, timeout=None):
        """
        Request removal of the roster entry identified by the given bare
        `jid`. If the entry currently has any subscription state, the server
        will send the corresponding unsubscribing presence stanzas.

        `timeout` is the maximum time in seconds to wait for a reply from the
        server.

        This may raise arbitrary :class:`.errors.XMPPError` exceptions if the
        server replies with an error and also any kind of connection error if
        the connection gets fatally terminated while waiting for a response.
        """
        yield from self.client.send(
            stanza.IQ(
                structs.IQType.SET,
                payload=roster_xso.Query(items=[
                    roster_xso.Item(
                        jid=jid,
                        subscription="remove"
                    )
                ])
            ),
            timeout=timeout
        )

    def approve(self, peer_jid):
        """
        (Pre-)approve a subscription request from `peer_jid`.

        :param peer_jid: The peer to (pre-)approve.

        This sends a ``"subscribed"`` presence to the peer; if the peer has
        previously asked for a subscription, this will seal the deal and create
        the subscription.

        If the peer has not requested a subscription (yet), it is marked as
        pre-approved by the server. A future subscription request by the peer
        will then be confirmed by the server automatically.

        .. note::

            Pre-approval is an OPTIONAL feature in :rfc:`6121`. It is announced
            as a stream feature.
        """
        self.client.enqueue(
            stanza.Presence(type_=structs.PresenceType.SUBSCRIBED,
                            to=peer_jid)
        )

    def subscribe(self, peer_jid):
        """
        Request presence subscription with the given `peer_jid`.

        This is deliberately not a coroutine; we don’t know whether the peer is
        online (usually) and they may defer the confirmation very long, if they
        confirm at all. Use :meth:`on_subscribed` to get notified when a peer
        accepted a subscription request.
        """
        self.client.enqueue(
            stanza.Presence(type_=structs.PresenceType.SUBSCRIBE,
                            to=peer_jid)
        )

    def unsubscribe(self, peer_jid):
        """
        Unsubscribe from the presence of the given `peer_jid`.
        """
        self.client.enqueue(
            stanza.Presence(type_=structs.PresenceType.UNSUBSCRIBE,
                            to=peer_jid)
        )
